/*******************************************************************************
* Copyright 2019-2024 Intel Corporation
*
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
*
*     http://www.apache.org/licenses/LICENSE-2.0
*
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*******************************************************************************/

#include <cassert>
#include <string>
#include <vector>
#include <unordered_map>
#ifdef DNNL_DEV_MODE
#include <fstream>
#include <iostream>
#include <sstream>
#endif

#include "gpu/intel/ocl/kernel_utils.hpp"

namespace dnnl {
namespace impl {
namespace gpu {
namespace intel {
namespace ocl {


extern const char *gen9_bnorm_bwd_kernel;
extern const char *gen9_bnorm_fwd_kernel;
extern const char *gen9_bnorm_nhwc_bwd_kernel;
extern const char *gen9_bnorm_nhwc_fwd_kernel;
extern const char *gen9_bnorm_reduce_kernel;
extern const char *nhwc_reusable_kernel;
extern const char *ref_bnorm_kernel;
extern const char *reusable_bnorm_kernel;
extern const char *simple_bnorm_kernel;
extern const char *custom_reorder_kernel;
extern const char *deconv_backward_bias_kernel;
extern const char *gemm_with_post_ops_kernel;
extern const char *ref_gemm_kernel;
extern const char *xe_hp_systolic_gemm_copy_kernel;
extern const char *xe_hpc_systolic_gemm_copy_kernel;
extern const char *gemm_post_ops_inner_product_kernel;
extern const char *gen9_binary_kernel;
extern const char *gen9_concat_kernel;
extern const char *gen9_eltwise_kernel;
extern const char *gen9_global_pooling_kernel;
extern const char *gen9_pooling_kernel;
extern const char *gen9_softmax_kernel;
extern const char *gen9_sum_kernel;
extern const char *gen9_wino_conv_fwd_data_2x3_kernel;
extern const char *gen9_wino_conv_fwd_data_fused_kernel;
extern const char *generic_reorder_kernel;
extern const char *many_inputs_sum_kernel;
extern const char *micro_sdpa_kernel;
extern const char *atomic_reduction_kernel;
extern const char *combined_reduction_kernel;
extern const char *ref_reduction_kernel;
extern const char *reusable_ref_reduction_kernel;
extern const char *ref_convolution_kernel;
extern const char *ref_eltwise_kernel;
extern const char *ref_group_normalization_kernel;
extern const char *ref_inner_product_kernel;
extern const char *ref_layer_normalization_kernel;
extern const char *ref_lrn_kernel;
extern const char *ref_matmul_kernel;
extern const char *ref_pooling_kernel;
extern const char *ref_prelu_kernel;
extern const char *ref_reorder_kernel;
extern const char *ref_resampling_kernel;
extern const char *ref_sdpa_kernel;
extern const char *ref_shuffle_kernel;
extern const char *ref_sparse_matmul_kernel;
extern const char *reusable_lnorm_kernel;
extern const char *reusable_simple_concat_kernel;
extern const char *reusable_softmax_kernel;
extern const char *reusable_vectorized_lnorm_kernel;
extern const char *rnn_grid_kernel;
extern const char *rnn_reorder_kernel;
extern const char *simple_binary_kernel;
extern const char *simple_concat_kernel;
extern const char *simple_layer_normalization_kernel;
extern const char *simple_softmax_kernel;
extern const char *simple_sum_kernel;
extern const char *simple_zero_pad_kernel;
extern const char *vectorized_lnorm_kernel;
extern const char *vectorized_lnorm_fused_kernel;
extern const char *vectorized_resampling_kernel;

extern const char *binary_common_header;
extern const char *gen9_bnorm_header;
extern const char *nhwc_reusable_header;
extern const char *concat_common_header;
extern const char *dispatch_header;
extern const char *ocl_gemm_attrs_header;
extern const char *layer_norm_common_header;
extern const char *metrics_discovery_api_header;
extern const char *ocl_conversion_header;
extern const char *ocl_custom_types_header;
extern const char *ocl_eltwise_header;
extern const char *ocl_generic_vector_ops_header;
extern const char *ocl_io_header;
extern const char *ocl_math_utils_header;
extern const char *ocl_post_ops_header;
extern const char *ocl_scales_header;
extern const char *ocl_types_header;
extern const char *ocl_utils_header;
extern const char *offsets_header;
extern const char *ocl_reduction_header;
extern const char *reorder_common_header;
extern const char *cell_compute_header;
extern const char *cell_kind_utility_header;
extern const char *rnn_common_header;
extern const char *sdpa_utils_header;
extern const char *simple_softmax_header;
extern const char *tile_ops_header;
extern const char *types_interop_header;

const char *get_kernel_source(const char *name) {
    static const std::unordered_map<std::string, const char *> kernel_list
            = { 
        { "gen9_calculate_stats", gen9_bnorm_bwd_kernel },
        { "gen9_bnorm_bwd", gen9_bnorm_bwd_kernel },
        { "gen9_calc_mean_var", gen9_bnorm_fwd_kernel },
        { "gen9_calc_mean", gen9_bnorm_fwd_kernel },
        { "gen9_calc_variance", gen9_bnorm_fwd_kernel },
        { "gen9_bnorm_fwd", gen9_bnorm_fwd_kernel },
        { "gen9_calculate_stats_nhwc", gen9_bnorm_nhwc_bwd_kernel },
        { "gen9_bnorm_bwd_nhwc", gen9_bnorm_nhwc_bwd_kernel },
        { "gen9_calc_mean_var_nhwc", gen9_bnorm_nhwc_fwd_kernel },
        { "gen9_calc_mean_nhwc", gen9_bnorm_nhwc_fwd_kernel },
        { "gen9_calc_variance_nhwc", gen9_bnorm_nhwc_fwd_kernel },
        { "gen9_bnorm_fwd_nhwc", gen9_bnorm_nhwc_fwd_kernel },
        { "gen9_fused_reduce_init", gen9_bnorm_reduce_kernel },
        { "gen9_fused_reduce_final", gen9_bnorm_reduce_kernel },
        { "gen9_fused_reduce_final", gen9_bnorm_reduce_kernel },
        { "gen9_reduce_mean_var", gen9_bnorm_reduce_kernel },
        { "gen9_reduce_mean", gen9_bnorm_reduce_kernel },
        { "gen9_reduce_variance", gen9_bnorm_reduce_kernel },
        { "gen9_reduce_stats", gen9_bnorm_reduce_kernel },
        { "nhwc_reusable_calc_mean", nhwc_reusable_kernel },
        { "nhwc_reusable_calc_mean_buff", nhwc_reusable_kernel },
        { "nhwc_reusable_calc_var", nhwc_reusable_kernel },
        { "nhwc_reusable_calc_var_buff", nhwc_reusable_kernel },
        { "nhwc_reusable_calc_mean_var", nhwc_reusable_kernel },
        { "nhwc_reusable_calc_mean_var_buff", nhwc_reusable_kernel },
        { "nhwc_reusable_norm_fwd", nhwc_reusable_kernel },
        { "nhwc_reusable_norm_fwd_buff", nhwc_reusable_kernel },
        { "nhwc_reusable_calc_stat", nhwc_reusable_kernel },
        { "nhwc_reusable_calc_stat_buff", nhwc_reusable_kernel },
        { "nhwc_reusable_norm_bwd", nhwc_reusable_kernel },
        { "nhwc_reusable_norm_bwd_buff", nhwc_reusable_kernel },
        { "nhwc_reusable_reduce_aux", nhwc_reusable_kernel },
        { "nhwc_reusable_reduce_fwd_reg", nhwc_reusable_kernel },
        { "nhwc_reusable_reduce_fwd_1pass", nhwc_reusable_kernel },
        { "nhwc_reusable_reduce_stat", nhwc_reusable_kernel },
        { "calculate_mean", ref_bnorm_kernel },
        { "calculate_variance", ref_bnorm_kernel },
        { "reduce_mean", ref_bnorm_kernel },
        { "reduce_variance", ref_bnorm_kernel },
        { "ref_bnorm_fwd", ref_bnorm_kernel },
        { "calculate_stats", ref_bnorm_kernel },
        { "reduce_stats", ref_bnorm_kernel },
        { "ref_bnorm_bwd", ref_bnorm_kernel },
        { "reusable_calculate_mean", reusable_bnorm_kernel },
        { "reusable_calculate_variance", reusable_bnorm_kernel },
        { "reusable_reduce_mean", reusable_bnorm_kernel },
        { "reusable_reduce_variance", reusable_bnorm_kernel },
        { "reusable_bnorm_fwd", reusable_bnorm_kernel },
        { "reusable_calculate_stats", reusable_bnorm_kernel },
        { "reusable_reduce_stats", reusable_bnorm_kernel },
        { "reusable_bnorm_bwd", reusable_bnorm_kernel },
        { "simple_calculate_mean_variance", simple_bnorm_kernel },
        { "simple_bnorm_fwd", simple_bnorm_kernel },
        { "simple_calculate_stats", simple_bnorm_kernel },
        { "simple_reduce_stats", simple_bnorm_kernel },
        { "simple_bnorm_bwd", simple_bnorm_kernel },
        { "custom_reorder", custom_reorder_kernel },
        { "deconv_backward_bias", deconv_backward_bias_kernel },
        { "gemm_post_ops", gemm_with_post_ops_kernel },
        { "ref_gemm", ref_gemm_kernel },
        { "xe_hp_systolic_gemm_copy", xe_hp_systolic_gemm_copy_kernel },
        { "xe_hp_systolic_gemm_copy", xe_hp_systolic_gemm_copy_kernel },
        { "xe_hp_systolic_gemm_copy", xe_hp_systolic_gemm_copy_kernel },
        { "xe_hp_systolic_gemm_copy", xe_hp_systolic_gemm_copy_kernel },
        { "xe_hp_systolic_gemm_copy", xe_hp_systolic_gemm_copy_kernel },
        { "xe_hp_systolic_gemm_copy", xe_hp_systolic_gemm_copy_kernel },
        { "xe_hpc_systolic_gemm_copy", xe_hpc_systolic_gemm_copy_kernel },
        { "xe_hpc_systolic_gemm_copy", xe_hpc_systolic_gemm_copy_kernel },
        { "xe_hpc_systolic_gemm_copy", xe_hpc_systolic_gemm_copy_kernel },
        { "xe_hpc_systolic_gemm_copy", xe_hpc_systolic_gemm_copy_kernel },
        { "xe_hpc_systolic_gemm_copy", xe_hpc_systolic_gemm_copy_kernel },
        { "xe_hpc_systolic_gemm_copy", xe_hpc_systolic_gemm_copy_kernel },
        { "gemm_post_ops_inner_product", gemm_post_ops_inner_product_kernel },
        { "gen9_binary", gen9_binary_kernel },
        { "gen9_binary", gen9_binary_kernel },
        { "gen9_binary", gen9_binary_kernel },
        { "gen9_binary", gen9_binary_kernel },
        { "gen9_concat", gen9_concat_kernel },
        { "gen9_eltwise_fwd", gen9_eltwise_kernel },
        { "gen9_eltwise_bwd", gen9_eltwise_kernel },
        { "gen9_global_pooling_fwd", gen9_global_pooling_kernel },
        { "gen9_global_pooling_bwd", gen9_global_pooling_kernel },
        { "gen9_pooling_fwd", gen9_pooling_kernel },
        { "gen9_pooling_bwd", gen9_pooling_kernel },
        { "gen9_softmax_fwd", gen9_softmax_kernel },
        { "gen9_softmax_bwd", gen9_softmax_kernel },
        { "gen9_sum", gen9_sum_kernel },
        { "gen9_wino_wei_transform_2x3", gen9_wino_conv_fwd_data_2x3_kernel },
        { "gen9_wino_src_transform_2x3", gen9_wino_conv_fwd_data_2x3_kernel },
        { "gen9_wino_dst_transform_2x3", gen9_wino_conv_fwd_data_2x3_kernel },
        { "gen9_wino_conv_fwd_2x3", gen9_wino_conv_fwd_data_2x3_kernel },
        { "gen9_wino_wei_transform", gen9_wino_conv_fwd_data_fused_kernel },
        { "gen9_wino_conv_fwd", gen9_wino_conv_fwd_data_fused_kernel },
        { "generic_reorder", generic_reorder_kernel },
        { "many_inputs_sum", many_inputs_sum_kernel },
        { "many_inputs_sum_batched", many_inputs_sum_kernel },
        { "micro_sdpa", micro_sdpa_kernel },
        { "atomic_reduce", atomic_reduction_kernel },
        { "combined_reduce", combined_reduction_kernel },
        { "ref_reduce", ref_reduction_kernel },
        { "reusable_ref_reduce", reusable_ref_reduction_kernel },
        { "ref_convolution_fwd", ref_convolution_kernel },
        { "ref_convolution_bwd_data", ref_convolution_kernel },
        { "ref_convolution_bwd_weights", ref_convolution_kernel },
        { "ref_eltwise_fwd", ref_eltwise_kernel },
        { "ref_eltwise_bwd", ref_eltwise_kernel },
        { "ref_gnorm_fwd", ref_group_normalization_kernel },
        { "ref_gnorm_bwd", ref_group_normalization_kernel },
        { "ref_inner_product_fwd", ref_inner_product_kernel },
        { "ref_inner_product_bwd_data", ref_inner_product_kernel },
        { "ref_inner_product_bwd_weights", ref_inner_product_kernel },
        { "ref_lnorm_fwd", ref_layer_normalization_kernel },
        { "ref_lnorm_bwd_scaleshift", ref_layer_normalization_kernel },
        { "ref_lnorm_bwd", ref_layer_normalization_kernel },
        { "ref_lrn_fwd", ref_lrn_kernel },
        { "ref_lrn_bwd", ref_lrn_kernel },
        { "ref_matmul", ref_matmul_kernel },
        { "ref_pooling_fwd", ref_pooling_kernel },
        { "ref_pooling_bwd", ref_pooling_kernel },
        { "ref_prelu_fwd", ref_prelu_kernel },
        { "ref_prelu_bwd", ref_prelu_kernel },
        { "subbyte_pack", ref_reorder_kernel },
        { "ref_reorder", ref_reorder_kernel },
        { "ref_resampling_fwd", ref_resampling_kernel },
        { "ref_resampling_bwd", ref_resampling_kernel },
        { "ref_sdpa", ref_sdpa_kernel },
        { "ref_shuffle", ref_shuffle_kernel },
        { "ref_sparse_matmul", ref_sparse_matmul_kernel },
        { "lnorm_reusable_calc_mean", reusable_lnorm_kernel },
        { "lnorm_reusable_calc_var", reusable_lnorm_kernel },
        { "lnorm_reusable_fwd", reusable_lnorm_kernel },
        { "lnorm_reusable_bwd_scaleshift", reusable_lnorm_kernel },
        { "lnorm_reusable_bwd", reusable_lnorm_kernel },
        { "reusable_simple_concat", reusable_simple_concat_kernel },
        { "reusable_softmax_fwd_generic", reusable_softmax_kernel },
        { "lnorm_reusable_vectorized", reusable_vectorized_lnorm_kernel },
        { "simple_rnn_copy_init_layer", rnn_grid_kernel },
        { "simple_rnn_copy_init_iter", rnn_grid_kernel },
        { "simple_rnn_copy_res_layer", rnn_grid_kernel },
        { "simple_rnn_copy_res_iter", rnn_grid_kernel },
        { "rnn_bias_prepare", rnn_grid_kernel },
        { "simple_rnn_elemwise_fwd", rnn_grid_kernel },
        { "simple_rnn_elemwise_fwd", rnn_grid_kernel },
        { "simple_rnn_elemwise_bwd", rnn_grid_kernel },
        { "simple_rnn_elemwise_bwd", rnn_grid_kernel },
        { "simple_rnn_cell_fwd", rnn_grid_kernel },
        { "simple_rnn_cell_fwd", rnn_grid_kernel },
        { "wei_reorder", rnn_reorder_kernel },
        { "simple_binary", simple_binary_kernel },
        { "simple_binary", simple_binary_kernel },
        { "simple_concat", simple_concat_kernel },
        { "simple_lnorm_fwd", simple_layer_normalization_kernel },
        { "simple_lnorm_bwd_scaleshift", simple_layer_normalization_kernel },
        { "simple_lnorm_bwd_scaleshift_final", simple_layer_normalization_kernel },
        { "simple_lnorm_bwd", simple_layer_normalization_kernel },
        { "simple_softmax_fwd_generic", simple_softmax_kernel },
        { "simple_softmax_bwd_generic", simple_softmax_kernel },
        { "simple_sum", simple_sum_kernel },
        { "simple_zero_pad", simple_zero_pad_kernel },
        { "simple_zero_pad_subg_16", simple_zero_pad_kernel },
        { "simple_zero_pad_subg_16_mask_and_clear_dt_1b", simple_zero_pad_kernel },
        { "vectorized_lnorm_fwd", vectorized_lnorm_kernel },
        { "vectorized_lnorm_bwd_scaleshift", vectorized_lnorm_kernel },
        { "vectorized_lnorm_bwd_scaleshift_final", vectorized_lnorm_kernel },
        { "vectorized_lnorm_bwd", vectorized_lnorm_kernel },
        { "vectorized_lnorm_bwd_fused", vectorized_lnorm_fused_kernel },
        { "vectorized_resampling_bwd", vectorized_resampling_kernel }, };

#ifdef DNNL_DEV_MODE
    static std::unordered_map<std::string, std::string> custom_kernels;
    const std::string runtime_kernel_overrides
            = dnnl::impl::gpu::intel::gpu_utils::dev_getenv(
                    "runtime_kernel_overrides", std::string());
    if (!runtime_kernel_overrides.empty()) {
        std::istringstream paths_stream(runtime_kernel_overrides);
        std::string name_path_pair;

        while (std::getline(paths_stream, name_path_pair, ':')) {
            std::size_t comma_pos = name_path_pair.find(',');
            if (comma_pos == std::string::npos) continue;

            std::string kernel_name = name_path_pair.substr(0, comma_pos);
            std::string kernel_path = name_path_pair.substr(comma_pos + 1);

            if (kernel_name != name) continue;

            auto it = custom_kernels.find(kernel_name);
            if (it == custom_kernels.end()) {
                VDEBUGINFO(1, common, kernel_list,
                        "loading kernel %s from file %s\n", kernel_name.c_str(),
                        kernel_path.c_str());
                std::ifstream file(kernel_path, std::ios::binary);
                gpu_assert(file) << "Could not open runtime_kernel_path file "
                                 << kernel_path << "\n";
                std::stringstream buffer;
                buffer << file.rdbuf();
                custom_kernels[kernel_name] = buffer.str();
            }

            return custom_kernels[kernel_name].c_str();
        }
    }
#endif

    if (!name) return nullptr;

    gpu_assert(kernel_list.count(name) == 1)
            << "Found " << kernel_list.count(name) << " kernels with the name "
            << name << ". Expected 1";
    return kernel_list.at(name);
}

const char *get_kernel_header(const std::string &name) {
    static const std::unordered_map<std::string, const char *>
            kernel_header_list
            = { 
        {"gpu/intel/ocl/binary_common.h", binary_common_header},
        {"gpu/intel/ocl/bnorm/gen9_bnorm.h", gen9_bnorm_header},
        {"gpu/intel/ocl/bnorm/nhwc_reusable.h", nhwc_reusable_header},
        {"gpu/intel/ocl/concat_common.h", concat_common_header},
        {"gpu/intel/ocl/dispatch.h", dispatch_header},
        {"gpu/intel/ocl/gemm/ocl_gemm_attrs.h", ocl_gemm_attrs_header},
        {"gpu/intel/ocl/layer_norm_common.h", layer_norm_common_header},
        {"gpu/intel/ocl/mdapi/metrics_discovery_api.h", metrics_discovery_api_header},
        {"gpu/intel/ocl/ocl_conversion.h", ocl_conversion_header},
        {"gpu/intel/ocl/ocl_custom_types.h", ocl_custom_types_header},
        {"gpu/intel/ocl/ocl_eltwise.h", ocl_eltwise_header},
        {"gpu/intel/ocl/ocl_generic_vector_ops.h", ocl_generic_vector_ops_header},
        {"gpu/intel/ocl/ocl_io.h", ocl_io_header},
        {"gpu/intel/ocl/ocl_math_utils.h", ocl_math_utils_header},
        {"gpu/intel/ocl/ocl_post_ops.h", ocl_post_ops_header},
        {"gpu/intel/ocl/ocl_scales.h", ocl_scales_header},
        {"gpu/intel/ocl/ocl_types.h", ocl_types_header},
        {"gpu/intel/ocl/ocl_utils.h", ocl_utils_header},
        {"gpu/intel/ocl/offsets.h", offsets_header},
        {"gpu/intel/ocl/reduction/ocl_reduction.h", ocl_reduction_header},
        {"gpu/intel/ocl/reorder_common.h", reorder_common_header},
        {"gpu/intel/ocl/rnn/cell_compute.h", cell_compute_header},
        {"gpu/intel/ocl/rnn/cell_kind_utility.h", cell_kind_utility_header},
        {"gpu/intel/ocl/rnn/rnn_common.h", rnn_common_header},
        {"gpu/intel/ocl/sdpa_utils.h", sdpa_utils_header},
        {"gpu/intel/ocl/simple_softmax.h", simple_softmax_header},
        {"gpu/intel/ocl/tile_ops.h", tile_ops_header},
        {"gpu/intel/ocl/types_interop.h", types_interop_header}, };

    gpu_assert(kernel_header_list.count(name) == 1)
            << "Found " << kernel_header_list.count(name)
            << " headers with the name " << name << ". Expected 1";
    return kernel_header_list.at(name);
}

} // namespace ocl
} // namespace intel
} // namespace gpu
} // namespace impl
} // namespace dnnl
